"""
lanpartydb.writing
~~~~~~~~~~~~~~~~~~

Data writing

:Copyright: 2024 Jochen Kupperschmidt
:License: MIT
"""

import dataclasses
from typing import Any

from lanpartydb.models import Party
import tomlkit


# party


def serialize_party(party: Party) -> str:
    """Serialize party to TOML document."""
    party_dict = _party_to_sparse_dict(party)

    location = party_dict.get('location', None)
    if location is not None:
        _convert_decimal_to_float(location, 'latitude')
        _convert_decimal_to_float(location, 'longitude')

    return _write_toml(party_dict)


def _party_to_sparse_dict(party: Party) -> dict[str, Any]:
    data = dataclasses.asdict(party)
    _remove_default_values(data)
    return data


def _convert_decimal_to_float(d: dict[str, Any], key: str) -> None:
    value = d.get(key)
    if value is not None:
        d[key] = float(value)


# util


def _write_toml(d: dict[str, Any]) -> str:
    return tomlkit.dumps(d)


def _remove_default_values(d: dict[str, Any]) -> dict[str, Any]:
    """Remove `None` and `False`, values from first level of dictionary."""
    for k, v in list(d.items()):
        if (v is None) or (v is False):
            del d[k]
        elif isinstance(v, dict):
            _remove_default_values(v)

    return d



"""
lanpartydb.models
~~~~~~~~~~~~~~~~~

Data models

:Copyright: 2024 Jochen Kupperschmidt
:License: MIT
"""

from __future__ import annotations
from dataclasses import dataclass, field
from datetime import date
from decimal import Decimal


@dataclass(frozen=True)
class Series:
    slug: str
    name: str
    alternative_names: list[str] = field(default_factory=list)
    country_codes: list[str] = field(default_factory=list)


@dataclass(frozen=True, kw_only=True)
class Party:
    slug: str
    title: str
    series_slug: str | None = field(kw_only=True, default=None)
    organizer_entity: str | None = field(kw_only=True, default=None)
    start_on: date
    end_on: date
    seats: int | None = None
    attendees: int | None = None
    online: bool | None = False
    location: Location | None = None
    links: Links | None = None


@dataclass(frozen=True)
class Location:
    name: str | None = field(kw_only=True, default=None)
    country_code: str
    city: str
    zip_code: str | None = None
    street: str | None = None
    latitude: Decimal | None = None
    longitude: Decimal | None = None


@dataclass(frozen=True)
class Links:
    website: Resource | None


@dataclass(frozen=True)
class Resource:
    url: str
    offline: bool = False



"""
lanpartydb.reading
~~~~~~~~~~~~~~~~~~

Data reading

:Copyright: 2024 Jochen Kupperschmidt
:License: MIT
"""

from decimal import Decimal
from pathlib import Path
import tomllib
from typing import Any

from .models import Links, Location, Party, Resource, Series


# series


def read_series_list_from_toml_file(filename: Path) -> list[Series]:
    """Read list of series from a TOML file."""
    toml = filename.read_text()
    return read_series_list_from_toml(toml)


def read_series_list_from_toml(toml: str) -> list[Series]:
    """Read list of series from a TOML document."""
    data = _load_toml(toml)
    return _read_series_list_from_dict(data)


def _read_series_list_from_dict(data: dict[str, Any]) -> list[Series]:
    """Read list of series from a dictionary."""
    return [Series(**item) for item in data.get('series', [])]


# party


def read_party_from_toml_file(filename: Path) -> Party:
    """Read party from a TOML file."""
    toml = filename.read_text()
    return read_party_from_toml(toml)


def read_party_from_toml(toml: str) -> Party:
    """Read party from a TOML document."""
    data = _load_toml(toml)
    return _read_party_from_dict(data)


def _read_party_from_dict(party_dict: dict[str, Any]) -> Party:
    """Read party from a dictionary."""
    location_dict = party_dict.pop('location', None)
    if location_dict:
        party_dict['location'] = Location(**location_dict)

    links_dict = party_dict.pop('links', None)
    if links_dict:
        website_dict = links_dict.pop('website', None)
        if website_dict:
            website = Resource(
                url=website_dict['url'],
                offline=website_dict.get('offline', False),
            )
            party_dict['links'] = Links(website=website)

    return Party(**party_dict)


# util


def _load_toml(toml: str) -> str:
    return tomllib.loads(toml, parse_float=Decimal)



"""
:Copyright: 2024 Jochen Kupperschmidt
:License: MIT
"""

import pytest

from lanpartydb.models import Series
from lanpartydb.reading import read_series_list_from_toml


@pytest.mark.parametrize(
    ('toml', 'expected'),
    [
        (
            """
            """,
            [],
        ),
        (
            """
            [[series]]
            slug = "megalan"
            name = "MegaLAN"
            """,
            [
                Series(
                    slug='megalan',
                    name='MegaLAN',
                    alternative_names=[],
                    country_codes=[],
                ),
            ],
        ),
        (
            """
            [[series]]
            slug = "gammalan"
            name = "GammaLAN"
            country_codes = ["ca", "us"]

            [[series]]
            slug = "deltalan"
            name = "DeltaLAN"
            alternative_names = ["Δ LAN", "Δέλτα LAN"]
            country_codes = ["au"]
            """,
            [
                Series(
                    slug='gammalan',
                    name='GammaLAN',
                    alternative_names=[],
                    country_codes=['ca', 'us'],
                ),
                Series(
                    slug='deltalan',
                    name='DeltaLAN',
                    alternative_names=['Δ LAN', 'Δέλτα LAN'],
                    country_codes=['au'],
                ),
            ],
        ),
    ],
)
def test_read_series_list_from_toml(toml: str, expected: list[Series]):
    assert read_series_list_from_toml(toml) == expected



"""
:Copyright: 2024 Jochen Kupperschmidt
:License: MIT
"""

from datetime import date
from decimal import Decimal
from textwrap import dedent

import pytest

from lanpartydb.models import Links, Location, Party, Resource
from lanpartydb.writing import serialize_party


@pytest.mark.parametrize(
    ('party', 'expected'),
    [
        (
            Party(
                slug='megalan-2023',
                title='MegaLAN 2023',
                start_on=date(2023, 11, 17),
                end_on=date(2023, 11, 19),
            ),
            dedent("""\
            slug = "megalan-2023"
            title = "MegaLAN 2023"
            start_on = 2023-11-17
            end_on = 2023-11-19
            """),
        ),
        (
            Party(
                slug='superlan-2024',
                title='SuperLAN 2024',
                series_slug='superlan',
                organizer_entity='SuperLAN Association',
                start_on=date(2024, 5, 24),
                end_on=date(2024, 5, 26),
                seats=420,
                attendees=397,
                online=False,
                location=Location(
                    name='City Hall',
                    country_code='us',
                    city='Los Angeles',
                    zip_code='90099',
                    street='123 North Hill Street',
                    latitude=Decimal('34.06101057935884'),
                    longitude=Decimal('-118.23974355902666'),
                ),
                links=Links(
                    website=Resource(
                        url='https://www.superlan.example/',
                        offline=False,
                    ),
                ),
            ),
            dedent("""\
            slug = "superlan-2024"
            title = "SuperLAN 2024"
            series_slug = "superlan"
            organizer_entity = "SuperLAN Association"
            start_on = 2024-05-24
            end_on = 2024-05-26
            seats = 420
            attendees = 397

            [location]
            name = "City Hall"
            country_code = "us"
            city = "Los Angeles"
            zip_code = "90099"
            street = "123 North Hill Street"
            latitude = 34.06101057935884
            longitude = -118.23974355902666

            [links.website]
            url = "https://www.superlan.example/"
            """),
        ),
    ],
)
def test_serialize_party(party: Party, expected: str):
    assert serialize_party(party) == expected
